import torch.nn as nn
import torch.optim as optim

import pytorch_lightning as pl
from argparse import ArgumentParser

from loguru import logger

from slp.util.log import configure_logging
from slp.config.config_parser import parse_config, make_cli_parser
from slp.plbind import (
    PLDataModuleFromDatasets,  # or PLDataModuleFromCorpus see slp/plbind/dm.py
    PLModule,  # or any other PLModule. See slp/plbind/module.py
    make_trainer,
    watch_model,
    FromLogits,
)


class MyCoolNet(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


# Configuration format looks like this:
# CONFIG_FORMAT = {
#     "seed": None,
#     "debug": False,  # Perform a debugging run
#     "data": {
#         # Args for PLDataModuleFromDatasets or PLDataModuleFromCorpus
#     }
#     "model": {
#         # Model parameters to be passed when model is instantiated.
#     },
#     "optimizer": "Adam",  # Or any other optimizer from torch.optim
#     "optim": {
#         "lr": 1e-3,
#         "weight_decay": 1e-2,
#     },
#     "trainer": {
#         "experiment_name": "my-cool-experiment",
#         # Args for make_trainer
#     },
#     "lr_scheduler": True,  # Default support for ReduceLROnPlateau
#     "lr_schedule": {
#         # Args for ReduceLROnPlateau
#     },
# }


def get_parser():
    parser = ArgumentParser("My cool experiment")
    # Add model args here. dest should be model.arg_name
    # For example
    # parser.add_argument(
    #     "--hidden",
    #     dest="model.intermediate_hidden",
    #     type=int,
    #     default=12,
    #     help="Intermediate hidden layers for linear module",
    # )
    return parser


def setup():
    # Your config parsing goes here.
    parser = get_parser()
    parser = make_cli_parser(parser, parser.parse_args().config)
    configure_logging(f"logs/{config.trainer.experiment_name}")
    # Add all default command line parsers and merge with yaml config file.
    config = parse_config(parser, PLDataModuleFromDatasets)
    return config


def get_data():
    # Download the original train, dev and test splits.
    # If a dev or test split is not provided, then it will be split randomly from the training set
    # Return torch.utils.data.Dataset to use with PLDataModuleFromDatasets or raw corpora to use with PLDataModuleFromCorpus
    train, val, test = None, None, None
    return train, val, test


def get_lightning_module(config):
    # Create your model, optimizer, criterion and lr_scheduler.
    model = MyCoolNet(**config.model)

    optimizer = getattr(optim, config.optimizer)(model.parameters(), **config.optim)

    lr_scheduler = None
    if config.lr_scheduler:
        lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, **config.lr_schedule
        )

    criterion = (
        nn.CrossEntropyLoss()
    )  # for classification or nn.MSELoss() for regression

    # Make your Lightning module. Note you can pass the metrics you need to monitor
    lm = PLModule(  # or RnnPLModule etc...
        model,
        optimizer,
        criterion,
        lr_scheduler=lr_scheduler,
        metrics={"acc": FromLogits(pl.metrics.classification.Accuracy())},
        hparams=config,  # This will automatically log configuration in wandb etc..
    )

    return lm


def get_lightning_data_module(config):
    # Get your data, preprocess, and create the LightningDataModule
    train, val, test = get_data()
    ldm = PLDataModuleFromDatasets(
        train, val=val, test=test, seed=config.seed, **config.data
    )
    return ldm


if __name__ == "__main__":
    # Boilerplate: Implement setup, get_data, get_lightning_data_module, get_lightning_module
    config = setup()

    if config.seed is not None:
        logger.info("Seeding everything with seed={seed}")
        pl.utilities.seed.seed_everything(seed=config.seed)

    ldm = get_lightning_data_module(config)
    lm = get_lightning_module(config)

    if config.debug:
        # Debug run on a small dataset

        logger.info("Running in debug mode: Fast run on 5 batches")
        trainer = make_trainer(fast_dev_run=5)
        trainer.fit(lm, datamodule=ldm)

        logger.info("Running in debug mode: Overfitting 5 batches")
        trainer = make_trainer(overfit_batches=5)
        trainer.fit(lm, datamodule=ldm)

    else:
        # Train and evaluate the model
        trainer = make_trainer(**config.trainer)
        watch_model(
            trainer, model
        )  # make wandb track gradients, parameters and model graph

        trainer.fit(lm, datamodule=ldm)

        trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())



import torch
import torch.nn as nn


class LayerNormTf(nn.Module):
    def __init__(self, hidden_size: int, eps: float = 1e-12):
        """Construct a layernorm module in the TF style (epsilon inside the square root).
        Link: https://github.com/huggingface/pytorch-pretrained-BERT/blob/master/pytorch_pretrained_bert/modeling.py#L234
        """
        super(LayerNormTf, self).__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size))
        self.bias = nn.Parameter(torch.zeros(hidden_size))
        self.variance_epsilon = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Calculate Layernorm the tf way"""
        u = x.mean(-1, keepdim=True)
        s = (x - u).pow(2).mean(-1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.variance_epsilon)
        return self.weight * x + self.bias


# Default to pytorch layernorm
LayerNorm = nn.LayerNorm



from slp.plbind.dm import PLDataModuleFromCorpus, PLDataModuleFromDatasets, split_data
from slp.plbind.helpers import EarlyStoppingWithLogs, FixedWandbLogger, FromLogits
from slp.plbind.module import (
    AutoEncoderPLModule,
    PLModule,
    RnnPLModule,
    SimplePLModule,
    TransformerClassificationPLModule,
    TransformerPLModule,
)
from slp.plbind.trainer import (
    add_optimizer_args,
    add_trainer_args,
    make_trainer,
    watch_model,
)



import argparse
from typing import IO, List, Optional, Union

import pytorch_lightning as pl
from loguru import logger
from slp.config.omegaconf import OmegaConfExtended as OmegaConf
from slp.plbind import add_optimizer_args, add_trainer_args


def make_cli_parser(
    parser: argparse.ArgumentParser, datamodule_cls: pl.LightningDataModule
):
    """make_cli_parser Augment an argument parser for slp with the default arguments

    Default arguments for training, logging, optimization etc. are added to the input {parser}.
    If you use make_cli_parser, the following command line arguments will be included

        usage: my_script.py [-h] [--optimizer {Adam,AdamW,SGD,Adadelta,Adagrad,Adamax,ASGD,RMSprop}] [--lr OPTIM.LR] [--weight-decay OPTIM.WEIGHT_DECAY]
                        [--lr-scheduler] [--lr-factor LR_SCHEDULE.FACTOR] [--lr-patience LR_SCHEDULE.PATIENCE] [--lr-cooldown LR_SCHEDULE.COOLDOWN] [--min-lr LR_SCHEDULE.MIN_LR] [--seed SEED]
                        [--config CONFIG] [--experiment-name TRAINER.EXPERIMENT_NAME] [--run-id TRAINER.RUN_ID] [--experiment-group TRAINER.EXPERIMENT_GROUP]
                        [--experiments-folder TRAINER.EXPERIMENTS_FOLDER] [--save-top-k TRAINER.SAVE_TOP_K] [--patience TRAINER.PATIENCE] [--wandb-project TRAINER.WANDB_PROJECT]
                        [--tags [TRAINER.TAGS [TRAINER.TAGS ...]]] [--stochastic_weight_avg] [--gpus TRAINER.GPUS] [--val-interval TRAINER.CHECK_VAL_EVERY_N_EPOCH]
                        [--clip-grad-norm TRAINER.GRADIENT_CLIP_VAL] [--epochs TRAINER.MAX_EPOCHS] [--steps TRAINER.MAX_STEPS] [--tbtt_steps TRAINER.TRUNCATED_BPTT_STEPS] [--debug]
                        [--val-percent DATA.VAL_PERCENT] [--test-percent DATA.TEST_PERCENT] [--bsz DATA.BATCH_SIZE] [--bsz-eval DATA.BATCH_SIZE_EVAL] [--num-workers DATA.NUM_WORKERS]
                        [--pin-memory] [--drop-last] [--shuffle-eval]

        optional arguments:
        -h, --help            show this help message and exit
        --optimizer {Adam,AdamW,SGD,Adadelta,Adagrad,Adamax,ASGD,RMSprop}
                                Which optimizer to use
        --lr OPTIM.LR         Learning rate
        --weight-decay OPTIM.WEIGHT_DECAY
                                Learning rate
        --lr-scheduler        Use learning rate scheduling. Currently only ReduceLROnPlateau is supported out of the box
        --lr-factor LR_SCHEDULE.FACTOR
                                Multiplicative factor by which LR is reduced. Used if --lr-scheduler is provided.
        --lr-patience LR_SCHEDULE.PATIENCE
                                Number of epochs with no improvement after which learning rate will be reduced. Used if --lr-scheduler is provided.
        --lr-cooldown LR_SCHEDULE.COOLDOWN
                                Number of epochs to wait before resuming normal operation after lr has been reduced. Used if --lr-scheduler is provided.
        --min-lr LR_SCHEDULE.MIN_LR
                                Minimum lr for LR scheduling. Used if --lr-scheduler is provided.
        --seed SEED           Seed for reproducibility
        --config CONFIG       Path to YAML configuration file
        --experiment-name TRAINER.EXPERIMENT_NAME
                                Name of the running experiment
        --run-id TRAINER.RUN_ID
                                Unique identifier for the current run. If not provided it is inferred from datetime.now()
        --experiment-group TRAINER.EXPERIMENT_GROUP
                                Group of current experiment. Useful when evaluating for different seeds / cross-validation etc.
        --experiments-folder TRAINER.EXPERIMENTS_FOLDER
                                Top-level folder where experiment results & checkpoints are saved
        --save-top-k TRAINER.SAVE_TOP_K
                                Save checkpoints for top k models
        --patience TRAINER.PATIENCE
                                Number of epochs to wait before early stopping
        --wandb-project TRAINER.WANDB_PROJECT
                                Wandb project under which results are saved
        --tags [TRAINER.TAGS [TRAINER.TAGS ...]]
                                Tags for current run to make results searchable.
        --stochastic_weight_avg
                                Use Stochastic weight averaging.
        --gpus TRAINER.GPUS   Number of GPUs to use
        --val-interval TRAINER.CHECK_VAL_EVERY_N_EPOCH
                                Run validation every n epochs
        --clip-grad-norm TRAINER.GRADIENT_CLIP_VAL
                                Clip gradients with ||grad(w)|| >= args.clip_grad_norm
        --epochs TRAINER.MAX_EPOCHS
                                Maximum number of training epochs
        --steps TRAINER.MAX_STEPS
                                Maximum number of training steps
        --tbtt_steps TRAINER.TRUNCATED_BPTT_STEPS
                                Truncated Back-propagation-through-time steps.
        --debug               If true, we run a full run on a small subset of the input data and overfit 10 training batches
        --val-percent DATA.VAL_PERCENT
                                Percent of validation data to be randomly split from the training set, if no validation set is provided
        --test-percent DATA.TEST_PERCENT
                                Percent of test data to be randomly split from the training set, if no test set is provided
        --bsz DATA.BATCH_SIZE
                                Training batch size
        --bsz-eval DATA.BATCH_SIZE_EVAL
                                Evaluation batch size
        --num-workers DATA.NUM_WORKERS
                                Number of workers to be used in the DataLoader
        --pin-memory          Pin data to GPU memory for faster data loading
        --drop-last           Drop last incomplete batch
        --shuffle-eval        Shuffle val & test sets
    Args:
        parser (argparse.ArgumentParser): A parent argument to be augmented
        datamodule_cls (pytorch_lightning.LightningDataModule): A data module class that injects arguments through the add_argparse_args method

    Returns:
        argparse.ArgumentParser: The augmented command line parser

    Examples:
        >>> import argparse
        >>> from slp.plbind.dm import PLDataModuleFromDatasets
        >>> parser = argparse.ArgumentParser("My cool model")
        >>> parser.add_argument("--hidden", dest="model.hidden", type=int)  # Create parser with model arguments and anything else you need
        >>> parser = make_cli_parser(parser, PLDataModuleFromDatasets)
        >>> args = parser.parse_args(args=["--bsz", "64", "--lr", "0.01"])
        >>> args.data.batch_size
        64
        >>> args.optim.lr
        0.01
    """
    parser = add_optimizer_args(parser)
    parser = add_trainer_args(parser)
    parser = datamodule_cls.add_argparse_args(parser)

    return parser


def parse_config(
    parser: argparse.ArgumentParser,
    config_file: Union[str, IO],
    args: Optional[List[str]] = None,
):
    """parse_config Parse a provided YAML config file and command line args and merge them

    During experimentation we want ideally to have a configuration file with the model and training configuration,
    but also be able to run quick experiments using command line args.
    This function allows you to double dip, by overriding values in a YAML config file through user provided command line arguments.

    The precedence for merging is as follows
       * default cli args values < config file values < user provided cli args

    E.g.:

       * if you don't include a value in your configuration it will take the default value from the argparse arguments
       * if you provide a cli arg (e.g. run the script with --bsz 64) it will override the value in the config file

    Note we use an extended OmegaConf istance to achieve this (see slp.config.omegaconf.OmegaConf)

    Args:
        parser (argparse.ArgumentParser): The argument parser you want to use
        config_file (Union[str, IO]): Configuration file name or file descriptor
        args (Optional[List[str]]): Optional input sys.argv style args. Useful for testing.
            Use this only for testing. By default it uses sys.argv[1:]

    Returns:
        OmegaConf.DictConfig: The parsed configuration as an OmegaConf DictConfig object

    Examples:
        >>> import io
        >>> from slp.config.config_parser import parse_config
        >>> mock_config_file = io.StringIO('''
        model:
          hidden: 100
        ''')
        >>> parser = argparse.ArgumentParser("My cool model")
        >>> parser.add_argument("--hidden", dest="model.hidden", type=int, default=20)
        >>> cfg = parse_config(parser, mock_config_file)
        {'model': {'hidden': 100}}
        >>> type(cfg)
        <class 'omegaconf.dictconfig.DictConfig'>
        >>> cfg = parse_config(parser, mock_config_file, args=["--hidden", "200"])
        {'model': {'hidden': 200}}
        >>> mock_config_file = io.StringIO('''
        random_value: hello
        ''')
        >>> cfg = parse_config(parser, mock_config_file)
        {'model': {'hidden': 20}, 'random_value': 'hello'}
    """
    # Merge Configurations Precedence: default kwarg values < default argparse values < config file values < user provided CLI args values
    if config_file is not None:
        dict_config = OmegaConf.from_yaml(config_file)  # type: ignore
    else:
        dict_config = OmegaConf.create({})

    user_cli, default_cli = OmegaConf.from_argparse(parser)
    config = OmegaConf.merge(default_cli, dict_config, user_cli)

    logger.info("Running with the following configuration")
    logger.info(f"\n{OmegaConf.to_yaml(config)}")

    return config



import pytorch_lightning as pl
import torch.nn as nn
from loguru import logger
from slp.data.collators import SequenceClassificationCollator
from slp.modules.classifier import Classifier
from slp.modules.rnn import WordRNN
from slp.plbind.dm import PLDataModuleFromCorpus
from slp.plbind.helpers import FromLogits
from slp.plbind.module import RnnPLModule
from slp.plbind.trainer import make_trainer, watch_model
from slp.util.log import configure_logging
from torch.optim import Adam
from torchnlp.datasets import smt_dataset  # type: ignore

collate_fn = SequenceClassificationCollator(device="cpu")


if __name__ == "__main__":
    EXPERIMENT_NAME = "smt-words-sentiment-classification"

    configure_logging(f"logs/{EXPERIMENT_NAME}")

    train, dev, test = smt_dataset(
        directory="../data/", train=True, dev=True, test=True
    )

    raw_train = [d["text"] for d in train]
    labels_train = [d["label"] for d in train]

    raw_dev = [d["text"] for d in dev]
    labels_dev = [d["label"] for d in dev]

    raw_test = [d["text"] for d in test]
    labels_test = [d["label"] for d in test]

    ldm = PLDataModuleFromCorpus(
        raw_train,
        labels_train,
        val=raw_dev,
        val_labels=labels_dev,
        test=raw_test,
        test_labels=labels_test,
        batch_size=8,
        batch_size_eval=32,
        collate_fn=collate_fn,
        pin_memory=True,
        num_workers=1,
        tokens="words",
        embeddings_file="./cache/glove.6B.50d.txt",
        embeddings_dim=50,
        lower=True,
        lang="en_core_web_md",
    )

    encoder = WordRNN(
        128,
        embeddings=ldm.embeddings,
        embeddings_dim=50,
        bidirectional=True,
        merge_bi="sum",
        packed_sequence=True,
        finetune_embeddings=True,
        attention=True,
    )

    model = Classifier(encoder, encoder.out_size, 3)

    optimizer = Adam([p for p in model.parameters() if p.requires_grad], lr=1e-3)
    criterion = nn.CrossEntropyLoss()

    lm = RnnPLModule(
        model,
        optimizer,
        criterion,
        metrics={"acc": FromLogits(pl.metrics.classification.Accuracy())},
    )

    trainer = make_trainer(
        EXPERIMENT_NAME,
        max_epochs=100,
        gpus=1,
        save_top_k=1,
    )
    watch_model(trainer, model)

    trainer.fit(lm, datamodule=ldm)

    trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slp', 'slp.config', 'slp.data', 'slp.modules', 'slp.plbind', 'slp.util']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'loguru>=0.5.3,<0.6.0',
 'mike>=0.6.0,<0.7.0',
 'nltk>=3.5,<4.0',
 'numpy>=1.20.1,<2.0.0',
 'omegaconf>=2.0.6,<3.0.0',
 'pytorch-ignite==0.2.0',
 'pytorch-lightning>=1.2.0,<2.0.0',
 'pytorch-nlp==0.4.1',
 'requests>=2.25.1,<3.0.0',
 'rpdb>=0.1.6,<0.2.0',
 'scikit-learn>=0.24.1,<0.25.0',
 'scipy>=1.6.1,<2.0.0',
 'sentencepiece>=0.1.95,<0.2.0',
 'spacy>=3.0.3,<4.0.0',
 'toml>=0.10.2,<0.11.0',
 'toolz>=0.11.1,<0.12.0',
 'torch>=1.7.1,<2.0.0',
 'torchvision>=0.8.2,<0.9.0',
 'tqdm>=4.57.0,<5.0.0',
 'transformers==4.3.0',
 'ujson>=4.0.2,<5.0.0',
 'validators>=0.18.2,<0.19.0',
 'wandb>=0.10.20,<0.11.0']

setup_kwargs = {
    'name': 'slp',
    'version': '1.1.1',
    'description': 'Speech, Language and Multimodal Processing models and utilities in PyTorch',
    'long_description': '# slp\n\n<p align="center">\n    <img src="https://github.com/georgepar/slp/actions/workflows/ci.yml/badge.svg" />\n    <img src="https://github.com/georgepar/slp/actions/workflows/docs.yml/badge.svg" />\n    <a href="https://codeclimate.com/github/georgepar/slp/maintainability" alt="Maintainability">\n        <img src="https://api.codeclimate.com/v1/badges/d3ad9729ad30aa158737/maintainability" /></a>\n    <a href="https://choosealicense.com/licenses/mit/" alt="License: MIT">\n        <img src="https://img.shields.io/badge/license-MIT-green.svg" /></a>\n    <a href="https://img.shields.io/pypi/pyversions/slp">\n        <img alt="Python Version" src="https://img.shields.io/pypi/pyversions/slp" /></a>\n    <a href="https://black.readthedocs.io/en/stable/" alt="Code Style: Black">\n        <img src="https://img.shields.io/badge/code%20style-black-000000.svg" /></a>\n</p>\n\n* **Repo:** [https://github.com/georgepar/slp](https://github.com/georgepar/slp)\n* **Documentation:** [https://georgepar.github.io/slp/latest/](https://georgepar.github.io/slp/latest/)\n\n\nslp is a framework for fast and reproducible development of multimodal models, with emphasis on\nNLP models.\n\nIt started as a collection of scripts and code I wrote / collected during my PhD and it evolves\naccordingly.\n\nAs such, the framework is opinionated and it follows a convention over configuration approach.\n\nA heavy emphasis is put on:\n\n- Enforcing best practices and reproducibility of experiments\n- Making common things fast at the top-level and not having to go through extensive configuration options\n- Remaining extendable. Extensions and modules for more use cases should be easy to add\n- Out of the box extensive logging and experiment management\n- Separating dirty / scratch code (at the script level) for quick changes and clean / polished code at the library level\n\nThis is currently in alpha release under active development, so things may break and new features\nwill be added.\n\n## Dependencies\n\nWe use [Pytorch](https://pytorch.org/) (1.7) and the following libraries\n\n- [Pytorch Lightning](https://pytorch-lightning.readthedocs.io/en/stable/)\n- [huggingface/transformers](https://huggingface.co/transformers/)\n- [Wandb](https://wandb.ai/)\n- Python 3.8\n\n## Installation\n\nYou can use slp as an external library by installing from PyPI with\n\n```\npip install slp\n```\n\nOr you can clone it from github\n\n```\ngit clone git@github.com:georgepar/slp\n```\n\nWe use [poetry](https://python-poetry.org/) for dependency management\n\nWhen you clone the repo run:\n\n```bash\npip install poetry\npoetry install\n```\n\nand a clean environment with all the dependencies will be created.\nYou can access it with `poetry shell`.\n\n**Note**: Wandb logging is enabled by default. You can either\n\n- Create an account and run `wandb login` when you clone the repo in a new machine to store the results in the online managed environment\n- Run `wandb offline` when you clone the repo to disable remote sync or use the `--offline` command\n  line argument in your scripts\n- Use one of their self-hosted solutions\n\n\n## Create a new project based on slp\n\nYou can use the template at [https://github.com/georgepar/cookiecutter-pytorch-slp](https://github.com/georgepar/cookiecutter-pytorch-slp)\nto create a new project based on slp\n\n```\npip install cookiecutter poetry\ncookiecutter gh:georgepar/cookiecutter-pytorch-slp\n# Follow the interactive configuration and a new folder with the project name you provided will appear\ncd $PROJECT_NAME\npoetry install  # Installs slp and all other dependencies\n```\n\nAnd you are good to go. Follow the instructions in the README of the new project you created. Happy coding\n\n## Contributing\n\nYou are welcome to open issues / PRs with improvements and bug fixes.\n\nSince this is mostly a personal project based around workflows and practices that work for me, I don\'t guarantee I will accept every change, but I\'m always open to discussion.\n\nIf you are going to contribute, please use the pre-commit hooks under `hooks`, otherwise the PR will not go through the CI. And never, ever touch `requirements.txt` by hand, it will automatically be exported from `poetry`\n\n```bash\n\ncat <<EOT >> .git/hooks/pre-commit\n#!/usr/bin/env bash\n\nbash hooks/black-check\nbash hooks/export-requirements-txt\nbash hooks/checks\nEOT\n\nchmod +x .git/hooks/pre-commit\n```\n\n## Cite\n\nIf you use this code for your research, please include the following citation\n\n```\n@ONLINE {,\n    author = "Georgios Paraskevopoulos",\n    title  = "slp",\n    year   = "2020",\n    url    = "https://github.com/georgepar/slp"\n}\n```\n\n\n## Roadmap\n\n* Optuna integration for hyperparameter tuning\n* Add dataloaders for popular multimodal datasets\n* Add multimodal architectures\n* Add RIM, DNC and Kanerva machine implementations\n* Write unit tests\n',
    'author': 'Giorgos Paraskevopoulos',
    'author_email': 'geopar@central.ntua.gr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://georgepar.github.io/slp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from argparse import Namespace
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import pytorch_lightning as pl
import torch
import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig
from slp.config.omegaconf import OmegaConf
from slp.util.pytorch import pad_mask, subsequent_mask
from slp.util.system import print_separator
from slp.util.types import Configuration, LossType
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler


class _Predictor(ABC):
    """Base predictor class

    Define an interface that can be used to extend the lightning module to new tasks and models

    * parse_batch: Parse input batch and extract necessery masks etc.
    * get_predictions_and_targets: Perform a forward pass through the model to get the logits and return logits and targets
    """

    @abstractmethod
    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Abstract parse_batch method to be implemented by child class

        Args:
            batch (Tuple[torch.Tensor, ...]): A tuple of tensors that contains inputs to the model and targets

        Returns:
            Tuple[torch.Tensor, ...]: The processed inputs, ready to provide to the model
        """
        pass

    @abstractmethod
    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Abstract get_predictions and targets method to be implemented by child class

        This method gets exposed to the PLModule classes

        Args:
            model (nn.Module): model to use for forward pass
            batch (Tuple[torch.Tensor, torch.Tensor]): A tuple of tensors that contains inputs to the model and targets

        **Note**: Maybe it should be useful to move loss calculation here. Then multitask learning and auxiliary losses should be easier

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, ground_truths), ready to be passed to the loss function
        """
        pass


class _Classification(_Predictor):
    """Classification task"""

    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Parse incoming batch

        Input batch just contains inputs and targets

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, labels)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (inputs, labels)
        """
        inputs = batch[0]
        targets = batch[1]

        return inputs, targets

    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return logits and ground truths to be passed in loss function

        Args:
            model (nn.Module): Model to use for prediction
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, labels)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, labels)
        """
        inputs, targets = self.parse_batch(batch)
        y_pred = model(inputs)

        return y_pred, targets


class _AutoEncoder(_Predictor):
    """Autoencoder task"""

    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Parse incoming batch

        Input batch just contains inputs. Targets are the same as inputs, because we are doing reconstruction.

        Args:
            batch (Tuple[torch.Tensor]): (inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (inputs, inputs)
        """
        inputs = batch[0]

        return inputs, inputs

    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return logits and ground truths to be passed in loss function

        Args:
            model (nn.Module): Model to use for prediction
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, inputs)
        """
        inputs, targets = self.parse_batch(batch)
        y_pred = model(inputs)

        return y_pred.view(y_pred.size(0), -1), targets.view(targets.size(0), -1)


class _RnnClassification(_Predictor):
    """RNN classification task"""

    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Parse incoming batch

        Input batch just contains inputs, targets and lengths.
        Comes from slp.data.collators.SequentialCollator.

        Args:
            batch (Tuple[torch.Tensor]): (inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (inputs, inputs)
        """
        inputs = batch[0]
        targets = batch[1]
        lengths = batch[2]

        return inputs, targets, lengths

    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return logits and ground truths to be passed in loss function

        Args:
            model (nn.Module): Model to use for prediction
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, inputs)
        """
        inputs, targets, lengths = self.parse_batch(batch)
        y_pred = model(inputs, lengths)

        return y_pred, targets


class _TransformerClassification(_Predictor):
    """Transformer classification task"""

    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Parse incoming batch

        Input batch just contains inputs, targets and lengths.
        Comes from slp.data.collators.SequentialCollator.
        Create pad masks to be passed to transformer attention

        Args:
            batch (Tuple[torch.Tensor]): (inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (inputs, inputs)
        """
        inputs = batch[0]
        targets = batch[1]
        lengths = batch[2]
        attention_mask = pad_mask(lengths)

        return inputs, targets, attention_mask

    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return logits and ground truths to be passed in loss function

        Args:
            model (nn.Module): Model to use for prediction
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, inputs)
        """
        inputs, targets, attention_mask = self.parse_batch(batch)
        y_pred = model(inputs, attention_mask=attention_mask)

        return y_pred, targets


class _Transformer(_Predictor):
    """Generic transformer seq2seq task"""

    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Parse incoming batch

        Input batch just contains inputs, targets and lengths.
        Comes from slp.data.collators.SequentialCollator.
        Create pad masks and subsequent_masks to be passed to transformer attention

        Args:
            batch (Tuple[torch.Tensor]): (inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (inputs, inputs)
        """
        inputs = batch[0]
        targets = batch[1]
        lengths_inputs = batch[2]
        lengths_targets = batch[3]

        max_length_inputs = torch.max(lengths_inputs)
        max_length_targets = torch.max(lengths_targets)

        pad_inputs = pad_mask(
            lengths_inputs,
            max_length=max_length_inputs,
        ).unsqueeze(-2)
        pad_targets = pad_mask(
            lengths_targets,
            max_length=max_length_targets,
        ).unsqueeze(-2)
        sub_m = subsequent_mask(max_length_targets)  # type: ignore
        pad_targets = pad_targets * sub_m.to(pad_targets.device)

        return inputs, targets, pad_inputs, pad_targets

    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return logits and ground truths to be passed in loss function

        Args:
            model (nn.Module): Model to use for prediction
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, inputs)
        """
        inputs, targets, source_mask, target_mask = self.parse_batch(batch)
        y_pred = model(
            inputs, targets, source_mask=source_mask, target_mask=target_mask
        )

        y_pred = y_pred.view(-1, y_pred.size(-1))
        targets = targets.view(-1)

        return y_pred, targets


class _BertSequenceClassification(_Predictor):
    """ Bert Classification task"""

    def parse_batch(self, batch: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        """Parse incoming batch

        Input batch just contains inputs, targets and lengths.
        Comes from slp.data.collators.SequentialCollator.
        Create pad masks to be passed to BERT attention

        Args:
            batch (Tuple[torch.Tensor]): (inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (inputs, inputs)
        """
        inputs = batch[0]
        targets = batch[1]
        lengths = batch[2]

        attention_mask = pad_mask(lengths)

        return inputs, targets, attention_mask

    def get_predictions_and_targets(
        self, model: nn.Module, batch: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Return logits and ground truths to be passed in loss function

        Args:
            model (nn.Module): Model to use for prediction
            batch (Tuple[torch.Tensor, torch.Tensor]): (inputs, inputs)

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (logits, inputs)
        """
        inputs, targets, attention_mask = self.parse_batch(batch)
        out = model(
            input_ids=inputs,
            attention_mask=attention_mask,
            labels=None,
            return_dict=False,
        )
        y_pred = out[0].view(-1, out[0].size(-1))
        targets = targets.view(-1)

        return y_pred, targets


class SimplePLModule(pl.LightningModule):
    def __init__(
        self,
        model: nn.Module,
        optimizer: Union[Optimizer, List[Optimizer]],
        criterion: LossType,
        lr_scheduler: Union[_LRScheduler, List[_LRScheduler]] = None,
        hparams: Configuration = None,
        metrics: Optional[Dict[str, pl.metrics.Metric]] = None,
        predictor_cls=_Classification,
        calculate_perplexity: bool = False,  # for LM. Dirty but much more efficient
    ):
        """LightningModule wrapper for a (model, optimizer, criterion, lr_scheduler) tuple

        Handles the boilerplate for metrics calculation and logging and defines the train_step / val_step / test_step
        with use of the predictor helper classes (e.g. _Classification, _RnnClassification)

        Args:
            model (nn.Module): Module to use for prediction
            optimizer (Union[Optimizer, List[Optimizer]]): Optimizers to use for training
            criterion (LossType): Task loss
            lr_scheduler (Union[_LRScheduler, List[_LRScheduler]], optional): Learning rate scheduler. Defaults to None.
            hparams (Configuration, optional): Hyperparameter values. This ensures they are logged with trainer.loggers. Defaults to None.
            metrics (Optional[Dict[str, pl.metrics.Metric]], optional): Metrics to track. Defaults to None.
            predictor_cls ([type], optional): Class that defines a parse_batch and a
                    get_predictions_and_targets method. Defaults to _Classification.
            calculate_perplexity (bool, optional): Whether to calculate perplexity.
                    Would be cleaner as a metric, but this is more efficient. Defaults to False.
        """
        super(SimplePLModule, self).__init__()
        self.calculate_perplexity = calculate_perplexity
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.criterion = criterion

        if metrics is not None:
            self.train_metrics = nn.ModuleDict(metrics)
            self.val_metrics = nn.ModuleDict({k: v.clone() for k, v in metrics.items()})
            self.test_metrics = nn.ModuleDict(
                {k: v.clone() for k, v in metrics.items()}
            )
        else:
            self.train_metrics = nn.ModuleDict(modules=None)
            self.val_metrics = nn.ModuleDict(modules=None)
            self.test_metrics = nn.ModuleDict(modules=None)
        self.predictor = predictor_cls()

        if hparams is not None:
            if isinstance(hparams, Namespace):
                dict_params = vars(hparams)
            elif isinstance(hparams, DictConfig):
                dict_params = cast(Dict[str, Any], OmegaConf.to_container(hparams))
            else:
                dict_params = hparams
            # self.hparams = dict_params
            self.save_hyperparameters(dict_params)

    def configure_optimizers(self):
        """Return optimizers and learning rate schedulers

        Returns:
            Tuple[List[Optimizer], List[_LRScheduler]]: (optimizers, lr_schedulers)
        """
        if self.lr_scheduler is not None:
            return self.optimizer, self.lr_scheduler
        else:
            return self.optimizer

    def forward(self, *args, **kwargs):
        """ Call wrapped module forward"""
        return self.model(*args, **kwargs)

    def _compute_metrics(self, metrics, loss, y_hat, targets, mode="train"):
        """Compute all metrics and aggregate in a dict

        Args:
            metrics (Dict[str, pl.metrics.Metric]): metrics to compute
            loss (torch.Tensor): Computed loss
            y_hat (torch.Tensor): Logits
            targets (torch.Tensor): Ground Truths
            mode (str, optional): "train", "val" or "test". Defaults to "train".
        """

        def fmt(name):
            """Format metric name"""
            return f"{mode}_{name}"

        metrics = {f"{mode}_{k}": v(y_hat, targets) for k, v in metrics.items()}

        if mode == "train":
            metrics["loss"] = loss
        else:
            metrics[fmt("loss")] = loss

        if self.calculate_perplexity:
            metrics[fmt("ppl")] = torch.exp(loss)

        return metrics

    def log_to_console(self, metrics, mode="Training"):
        """Log metrics to console

        Args:
            metrics (Dict[str, torch.Tensor]): Computed metrics
            mode (str, optional): "Training", "Validation" or "Testing". Defaults to "Training".
        """
        logger.info("Epoch {} {} results".format(self.current_epoch + 1, mode))
        print_separator(symbol="-", n=50, print_fn=logger.info)

        for name, value in metrics.items():
            if name == "epoch":
                continue
            logger.info("{:<15} {:<15}".format(name, value))

        print_separator(symbol="%", n=50, print_fn=logger.info)

    def aggregate_epoch_metrics(self, outputs, mode="Training"):
        """Aggregate metrics over a whole epoch

        Args:
            outputs (List[Dict[str, torch.Tensor]]): Aggregated outputs from train_step, validation_step or test_step
            mode (str, optional): "Training", "Validation" or "Testing". Defaults to "Training".
        """

        def fmt(name):
            """Format metric name"""
            return f"{name}" if name != "loss" else "train_loss"

        keys = list(outputs[0].keys())
        aggregated = {fmt(k): torch.stack([x[k] for x in outputs]).mean() for k in keys}
        self.log_to_console(aggregated, mode=mode)
        aggregated["epoch"] = self.current_epoch + 1
        self.log_dict(aggregated, logger=True, prog_bar=True, on_epoch=True)

        return aggregated

    def training_step(self, batch, batch_idx):
        """Compute loss for a single training step and log metrics to loggers

        Args:
            batch (Tuple[torch.Tensor, ...]): Input batch
            batch_idx (int): Index of batch

        Returns:
            Dict[str, torch.Tensor]: computed metrics
        """
        y_hat, targets = self.predictor.get_predictions_and_targets(self.model, batch)
        loss = self.criterion(y_hat, targets)
        metrics = self._compute_metrics(
            self.train_metrics, loss, y_hat, targets, mode="train"
        )

        self.log_dict(
            {k: v for k, v in metrics.items()},
            on_step=True,
            on_epoch=False,
            logger=True,
            prog_bar=False,
        )

        metrics["loss"] = loss

        return metrics

    def training_epoch_end(self, outputs):
        """Aggregate metrics of a training epoch

        Args:
            outputs (List[Dict[str, torch.Tensor]]): Aggregated outputs from train_step
        """
        outputs = self.aggregate_epoch_metrics(outputs, mode="Training")

    def validation_step(self, batch, batch_idx):
        """Compute loss for a single validation step and log metrics to loggers

        Args:
            batch (Tuple[torch.Tensor, ...]): Input batch
            batch_idx (int): Index of batch

        Returns:
            Dict[str, torch.Tensor]: computed metrics
        """
        y_hat, targets = self.predictor.get_predictions_and_targets(self, batch)
        loss = self.criterion(y_hat, targets)
        metrics = self._compute_metrics(
            self.val_metrics, loss, y_hat, targets, mode="val"
        )

        return metrics

    def validation_epoch_end(self, outputs):
        """Aggregate metrics of a validation epoch

        Args:
            outputs (List[Dict[str, torch.Tensor]]): Aggregated outputs from validation_step
        """
        outputs = self.aggregate_epoch_metrics(outputs, mode="Validation")

    def test_step(self, batch, batch_idx):
        """Compute loss for a single test step and log metrics to loggers

        Args:
            batch (Tuple[torch.Tensor, ...]): Input batch
            batch_idx (int): Index of batch

        Returns:
            Dict[str, torch.Tensor]: computed metrics
        """
        y_hat, targets = self.predictor.get_predictions_and_targets(self, batch)
        loss = self.criterion(y_hat, targets)
        metrics = self._compute_metrics(
            self.test_metrics, loss, y_hat, targets, mode="test"
        )

        return metrics

    def test_epoch_end(self, outputs):
        """Aggregate metrics of a test epoch

        Args:
            outputs (List[Dict[str, torch.Tensor]]): Aggregated outputs from test_step
        """
        outputs = self.aggregate_epoch_metrics(outputs, mode="Test")


def _make_specialized_pl_module(predictor_cls):
    """Create a LightningModule wrapper using the provided predictor class

    Args:
        predictor_cls: Class that defines parse_batch and get_predictions_and_targets

    Returns:
        pl.LightningModule: Configured LightningModule
    """

    class Module(SimplePLModule):
        def __init__(
            self,
            model: nn.Module,
            optimizer: Union[Optimizer, List[Optimizer]],
            criterion: LossType,
            lr_scheduler: Union[_LRScheduler, List[_LRScheduler]] = None,
            hparams: Configuration = None,
            metrics: Optional[Dict[str, pl.metrics.Metric]] = None,
            calculate_perplexity=False,
        ):
            """Pass arguments through to base class"""
            super(Module, self).__init__(
                model,
                optimizer,
                criterion,
                predictor_cls=predictor_cls,
                lr_scheduler=lr_scheduler,
                hparams=hparams,
                metrics=metrics,
                calculate_perplexity=calculate_perplexity,
            )

    return Module


PLModule = _make_specialized_pl_module(_Classification)
AutoEncoderPLModule = _make_specialized_pl_module(_AutoEncoder)
RnnPLModule = _make_specialized_pl_module(_RnnClassification)
TransformerClassificationPLModule = _make_specialized_pl_module(
    _TransformerClassification
)
TransformerPLModule = _make_specialized_pl_module(_Transformer)
BertPLModule = _make_specialized_pl_module(_BertSequenceClassification)



from argparse import ArgumentParser

import pytorch_lightning as pl
import torch.nn as nn
from loguru import logger
from slp.config.config_parser import make_cli_parser, parse_config
from slp.data.collators import SequenceClassificationCollator
from slp.modules.classifier import Classifier
from slp.modules.rnn import WordRNN
from slp.plbind.dm import PLDataModuleFromCorpus
from slp.plbind.helpers import FromLogits
from slp.plbind.module import BertPLModule
from slp.plbind.trainer import make_trainer, watch_model
from slp.util.log import configure_logging
from torch.optim import Adam
from torchnlp.datasets import smt_dataset  # type: ignore
from transformers import AdamW, BertForSequenceClassification

collate_fn = SequenceClassificationCollator(device="cpu")


def get_parser():
    parser = ArgumentParser("Bert finetuning for SST-2")
    parser.add_argument(
        "--binary",
        dest="binary",
        action="store_true",
        help="Perform binary classification. If False performs fine-grained 5-class classification",
    )
    return parser


def get_data(config):
    train, dev, test = smt_dataset(
        directory="../data/",
        train=True,
        dev=True,
        test=True,
        fine_grained=True,
    )

    def filter_neutrals(data, labels):
        logger.info("Filtering neutral labels for binary task")

        new_data, new_labels = [], []

        for d, l in zip(data, labels):
            # l positive or very positive
            if "positive" in l:
                new_data.append(d)
                new_labels.append("positive")
            # l negative or very negative
            elif "negative" in l:
                new_data.append(d)
                new_labels.append("negative")
            else:
                continue

        return new_data, new_labels

    raw_train = [d["text"] for d in train]
    labels_train = [d["label"] for d in train]

    raw_dev = [d["text"] for d in dev]
    labels_dev = [d["label"] for d in dev]

    raw_test = [d["text"] for d in dev]
    labels_test = [d["label"] for d in dev]

    num_labels = 5
    if config.binary:
        raw_train, labels_train = filter_neutrals(raw_train, labels_train)
        raw_dev, labels_dev = filter_neutrals(raw_dev, labels_dev)
        raw_test, labels_test = filter_neutrals(raw_test, labels_test)
        num_labels = 2

    return (
        raw_train,
        labels_train,
        raw_dev,
        labels_dev,
        raw_test,
        labels_test,
        num_labels,
    )


if __name__ == "__main__":
    parser = get_parser()
    parser = make_cli_parser(parser, PLDataModuleFromCorpus)

    args = parser.parse_args()
    config_file = args.config

    config = parse_config(parser, config_file)
    # Set these by default.
    config.hugging_face_model = config.data.tokenizer
    config.data.add_special_tokens = True
    config.data.lower = "uncased" in config.hugging_face_model

    if config.trainer.experiment_name == "experiment":
        config.trainer.experiment_name = "finetune-bert-smt"

    configure_logging(f"logs/{config.trainer.experiment_name}")

    if config.seed is not None:
        logger.info("Seeding everything with seed={seed}")
        pl.utilities.seed.seed_everything(seed=config.seed)

    (
        raw_train,
        labels_train,
        raw_dev,
        labels_dev,
        raw_test,
        labels_test,
        num_labels,
    ) = get_data(config)

    ldm = PLDataModuleFromCorpus(
        raw_train,
        labels_train,
        val=raw_dev,
        val_labels=labels_dev,
        test=raw_test,
        test_labels=labels_test,
        collate_fn=collate_fn,
        **config.data,
    )

    model = BertForSequenceClassification.from_pretrained(
        config.hugging_face_model, num_labels=num_labels
    )

    logger.info(model)

    # Leave this hardcoded for now.
    optimizer = AdamW([p for p in model.parameters() if p.requires_grad], lr=1e-5)
    criterion = nn.CrossEntropyLoss()

    lm = BertPLModule(
        model,
        optimizer,
        criterion,
        metrics={"acc": FromLogits(pl.metrics.classification.Accuracy())},
    )

    trainer = make_trainer(**config.trainer)
    watch_model(trainer, model)

    trainer.fit(lm, datamodule=ldm)

    trainer.test(ckpt_path="best", test_dataloaders=ldm.test_dataloader())



import io
import logging
from typing import Any, Optional

from loguru import logger
from slp.util.system import date_fname
from tqdm import tqdm


def log_to_file(fname_prefix: Optional[str]) -> str:
    """log_to_file Configure loguru to log to a logfile

    Args:
        logfile_prefix (Optional[str]): Optional prefix to file where logs will be written.

    Returns:
        str: The logfile where logs are written
    """
    logfile = f"{fname_prefix}.{date_fname()}.log"
    logger.add(
        logfile,
        colorize=False,
        level="DEBUG",
        enqueue=True,
    )
    return logfile


def configure_logging(logfile_prefix: Optional[str] = None) -> str:
    """configure_logging Configure loguru to intercept logging module logs, tqdm.writes and write to a logfile

    We use logure for stdout/stderr logging in this project.
    This function configures loguru to intercept logs from other modules that use the default python logging module.
    It also configures loguru so that it plays well with writes in the tqdm progress bars
    If a logfile_prefix is provided, loguru will also write all logs into a logfile with a unique name constructed using
    logfile_prefix and datetime.now()

    Args:
        logfile_prefix (Optional[str]): Optional prefix to file where logs will be written.

    Returns:
        str: The logfile where logs are written

    Examples:
        >>> configure_logging("logs/my-cool-experiment)
        logs/my-cool-experiment.20210228-211832.log
    """

    class InterceptHandler(logging.Handler):
        def emit(self, record):
            """Intercept standard logging logs in loguru. Should test this for distributed pytorch lightning"""
            # Get corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where originated the logged message
            frame, depth = logging.currentframe(), 2
            while frame.f_code.co_filename == logging.__file__:
                frame = frame.f_back
                depth += 1

            logger.opt(depth=depth, exception=record.exc_info).log(
                level, record.getMessage()
            )

    logger.info("Intercepting standard logging logs in loguru")

    # Make loguru play well with tqdm
    logger.remove()

    def tqdm_write(msg: str) -> Any:
        """tqdm write wrapper for loguru"""
        return tqdm.write(msg, end="")

    logger.add(tqdm_write, colorize=True)

    logging.basicConfig(handlers=[InterceptHandler()], level=logging.INFO)

    if logfile_prefix is not None:
        logfile = log_to_file(logfile_prefix)
        logger.info(f"Log file will be saved in {logfile}")

    return logfile



from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from loguru import logger
from slp.modules.attention import Attention
from slp.modules.embed import Embed
from slp.util.pytorch import PackSequence, PadPackedSequence, pad_mask


class RNN(nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        batch_first: bool = True,
        layers: int = 1,
        bidirectional: bool = False,
        merge_bi: str = "cat",
        dropout: float = 0.0,
        rnn_type: str = "lstm",
        packed_sequence: bool = True,
    ):
        """LSTM - GRU wrapper with packed sequence support and handling for bidirectional / last output states

        It is recommended to run with batch_first=True because the rest of the code is built with this assumption

        Args:
            input_size (int): Input features.
            hidden_size (int): Hidden features.
            batch_first (bool): Use batch first representation type. Defaults to True.
            layers (int): Number of RNN layers. Defaults to 1.
            bidirectional (bool): Use bidirectional RNNs. Defaults to False.
            merge_bi (str): How bidirectional states are merged. Defaults to "cat".
            dropout (float): Dropout probability. Defaults to 0.0.
            rnn_type (str): lstm or gru. Defaults to "lstm".
            packed_sequence (bool): Use packed sequences. Defaults to True.
        """
        super(RNN, self).__init__()
        self.bidirectional = bidirectional
        self.hidden_size = hidden_size
        self.batch_first = batch_first
        self.merge_bi = merge_bi
        self.rnn_type = rnn_type.lower()

        if not batch_first:
            logger.warning(
                "You are running RNN with batch_first=False. Make sure this is really what you want"
            )

        if not packed_sequence:
            logger.warning(
                "You have set packed_sequence=False. Running with packed_sequence=True will be much faster"
            )

        rnn_cls = nn.LSTM if self.rnn_type == "lstm" else nn.GRU
        self.rnn = rnn_cls(
            input_size,
            hidden_size,
            batch_first=batch_first,
            num_layers=layers,
            bidirectional=bidirectional,
        )
        self.drop = nn.Dropout(dropout)
        self.packed_sequence = packed_sequence

        if packed_sequence:
            self.pack = PackSequence(batch_first=batch_first)
            self.unpack = PadPackedSequence(batch_first=batch_first)

    @property
    def out_size(cls) -> int:
        """RNN output features size

        Returns:
            int: RNN output features size
        """
        out: int = (
            2 * cls.hidden_size
            if cls.bidirectional and cls.merge_bi == "cat"
            else cls.hidden_size
        )
        return out

    def _merge_bi(self, forward: torch.Tensor, backward: torch.Tensor) -> torch.Tensor:
        """Merge forward and backward states

        Args:
            forward (torch.Tensor): [B, L, H] Forward states
            backward (torch.Tensor): [B, L, H] Backward states

        Returns:
            torch.Tensor: [B, L, H] or [B, L, 2*H] Merged forward and backward states
        """
        if self.merge_bi == "sum":
            return forward + backward

        return torch.cat((forward, backward), dim=-1)

    def _select_last_unpadded(
        self, out: torch.Tensor, lengths: torch.Tensor
    ) -> torch.Tensor:
        """Get the last timestep before padding starts

        Args:
            out (torch.Tensor): [B, L, H] Fprward states
            lengths (torch.Tensor): [B] Original sequence lengths

        Returns:
            torch.Tensor: [B, H] Features for last sequence timestep
        """
        gather_dim = 1 if self.batch_first else 0
        gather_idx = (
            (lengths - 1)  # -1 to convert to indices
            .unsqueeze(1)  # (B) -> (B, 1)
            .expand((-1, self.hidden_size))  # (B, 1) -> (B, H)
            # (B, 1, H) if batch_first else (1, B, H)
            .unsqueeze(gather_dim)
        )
        # Last forward for real length or seq (unpadded tokens)
        last_out = out.gather(gather_dim, gather_idx).squeeze(gather_dim)

        return last_out

    def _final_output(
        self, out: torch.Tensor, lengths: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Create RNN ouputs

        Collect last hidden state for forward and backward states
        Code adapted from https://stackoverflow.com/a/50950188

        Args:
            out (torch.Tensor): [B, L, num_directions * H] RNN outputs
            lengths (torch.Tensor): [B] Original sequence lengths

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: (
                merged forward and backward states [B, L, H] or [B, L, 2*H],
                merged last forward and backward state [B, H] or [B, 2*H]
            )
        """
        if not self.bidirectional:
            return out, self._select_last_unpadded(out, lengths)

        forward, backward = (out[..., : self.hidden_size], out[..., self.hidden_size :])
        # Last backward corresponds to first token
        last_backward_out = backward[:, 0, :] if self.batch_first else backward[0, ...]
        # Last forward for real length or seq (unpadded tokens)
        last_forward_out = self._select_last_unpadded(forward, lengths)
        out = self._merge_bi(forward, backward) if self.merge_bi != "cat" else out

        return out, self._merge_bi(last_forward_out, last_backward_out)

    def forward(
        self, x: torch.Tensor, lengths: torch.Tensor
    ) -> Tuple[
        torch.Tensor,
        torch.Tensor,
        Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]],
    ]:
        """RNN forward pass

        Args:
            x (torch.Tensor): [B, L, D] Input features
            lengths (torch.Tensor): [B] Original sequence lengths

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: (
                merged forward and backward states [B, L, H] or [B, L, 2*H],
                merged last forward and backward state [B, H] or [B, 2*H],
                hidden states tuple of [num_layers * num_directions, B, H] for LSTM or tensor [num_layers * num_directions, B, H] for GRU
            )
        """
        self.rnn.flatten_parameters()

        if self.packed_sequence:
            # Latest pytorch allows only cpu tensors for packed sequence
            lengths = lengths.to("cpu")
            x, lengths = self.pack(x, lengths)
        out, hidden = self.rnn(x)

        if self.packed_sequence:
            out = self.unpack(out, lengths)
        out = self.drop(out)
        lengths = lengths.to(out.device)

        out, last_timestep = self._final_output(out, lengths)

        return out, last_timestep, hidden


class WordRNN(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        vocab_size: Optional[int] = None,
        embeddings_dim: Optional[int] = None,
        embeddings: Optional[np.ndarray] = None,
        embeddings_dropout: float = 0.0,
        finetune_embeddings: bool = False,
        batch_first: bool = True,
        layers: int = 1,
        bidirectional: bool = False,
        merge_bi: str = "cat",
        dropout: float = 0.1,
        rnn_type: str = "lstm",
        packed_sequence: bool = True,
        attention: bool = False,
    ):
        """RNN with embedding layer and optional attention mechanism

        Single-headed scaled dot-product attention is used as an attention mechanism

        Args:
            hidden_size (int): Hidden features
            vocab_size (Optional[int]): Vocabulary size. Defaults to None.
            embeddings_dim (Optional[int]): Embedding dimension. Defaults to None.
            embeddings (Optional[np.ndarray]): Embedding matrix. Defaults to None.
            embeddings_dropout (float): Embedding dropout probability. Defaults to 0.0.
            finetune_embeddings (bool): Finetune embeddings? Defaults to False.
            batch_first (bool): Use batch first representation type. Defaults to True.
            layers (int): Number of RNN layers. Defaults to 1.
            bidirectional (bool): Use bidirectional RNNs. Defaults to False.
            merge_bi (str): How bidirectional states are merged. Defaults to "cat".
            dropout (float): Dropout probability. Defaults to 0.0.
            rnn_type (str): lstm or gru. Defaults to "lstm".
            packed_sequence (bool): Use packed sequences. Defaults to True.
            attention (bool): Use attention mechanism on RNN outputs. Defaults to False.
        """
        super(WordRNN, self).__init__()

        if embeddings is None:
            finetune_embeddings = True
            assert (
                vocab_size is not None
            ), "You should either pass an embeddings matrix or vocab size"
            assert (
                embeddings_dim is not None
            ), "You should either pass an embeddings matrix or embeddings_dim"
        else:
            vocab_size = embeddings.shape[0]
            embeddings_dim = embeddings.shape[1]
        self.embed = Embed(
            vocab_size,
            embeddings_dim,
            embeddings=embeddings,
            dropout=embeddings_dropout,
            trainable=finetune_embeddings,
        )
        self.rnn = RNN(
            embeddings_dim,
            hidden_size,
            batch_first=batch_first,
            layers=layers,
            merge_bi=merge_bi,
            bidirectional=bidirectional,
            dropout=dropout,
            rnn_type=rnn_type,
            packed_sequence=packed_sequence,
        )
        self.out_size = (
            hidden_size
            if not (bidirectional and merge_bi == "cat")
            else 2 * hidden_size
        )
        self.attention = None

        if attention:
            self.attention = Attention(attention_size=self.out_size, dropout=dropout)

    def forward(self, x: torch.Tensor, lengths: torch.Tensor) -> torch.Tensor:
        """Word RNN forward pass

        If self.attention=True then the outputs are the weighted sum of the RNN hidden states with the attention score weights
        Else the output is the last hidden state of the RNN.

        Args:
            x (torch.Tensor): [B, L] Input token ids
            lengths (torch.Tensor): [B] Original sequence lengths

        Returns:
            torch.Tensor: [B, H] or [B, 2*H] Output features to be used for classification
        """
        x = self.embed(x)
        out, last_hidden, _ = self.rnn(x, lengths)

        if self.attention is not None:
            out, _ = self.attention(out, attention_mask=pad_mask(lengths))
            out = out.sum(1)
        else:
            out = last_hidden

        return out  # type: ignore
